# Copyright 2016 OpenStack Foundation
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import extensions
from neutron_lib.db import constants as db_const

from neutron.extensions import stdattrs_common


DESCRIPTION_BODY = {
    'description': {'allow_post': True, 'allow_put': True,
                    'validate': {
                        'type:string': db_const.DESCRIPTION_FIELD_SIZE},
                    'is_visible': True, 'default': '', 'is_filter': True}
}


class Standardattrdescription(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "standard-attr-description"

    @classmethod
    def get_alias(cls):
        return "standard-attr-description"

    @classmethod
    def get_description(cls):
        return "Extension to add descriptions to standard attributes"

    @classmethod
    def get_updated(cls):
        return "2016-02-10T10:00:00-00:00"

    def get_optional_extensions(self):
        return ['security-group', 'router']

    def get_extended_resources(self, version):
        if version != "2.0":
            return {}
        return stdattrs_common.stdattrs_extended_resources(DESCRIPTION_BODY)



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import queue
from unittest import mock

import eventlet
from keystoneauth1 import exceptions as ks_exc
from neutron_lib import constants as n_const
from neutron_lib import context as n_ctx
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from novaclient import api_versions
from novaclient import exceptions as nova_exceptions
from oslo_config import cfg
from oslo_utils import uuidutils
from sqlalchemy.orm import attributes as sql_attr

from neutron.notifiers import nova
from neutron.objects import ports as port_obj
from neutron.tests import base

DEVICE_OWNER_COMPUTE = n_const.DEVICE_OWNER_COMPUTE_PREFIX + 'fake'
DEVICE_OWNER_BAREMETAL = n_const.DEVICE_OWNER_BAREMETAL_PREFIX + 'fake'


class TestNovaNotify(base.BaseTestCase):
    def setUp(self, plugin=None):
        super(TestNovaNotify, self).setUp()
        self.ctx = n_ctx.get_admin_context()
        self.port_uuid = uuidutils.generate_uuid()

        class FakePlugin(object):
            def get_port(self, context, port_id):
                device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
                return {'device_id': device_id,
                        'device_owner': DEVICE_OWNER_COMPUTE,
                        'id': port_id}

        self.nova_notifier = nova.Notifier()
        directory.add_plugin(plugin_constants.CORE, FakePlugin())

    def test_notify_port_status_all_values(self):
        states = [n_const.PORT_STATUS_ACTIVE, n_const.PORT_STATUS_DOWN,
                  n_const.PORT_STATUS_ERROR, n_const.PORT_STATUS_BUILD,
                  None]
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        # test all combinations
        for previous_port_status in states:
            for current_port_status in states:
                params = {'id': self.port_uuid, 'device_id': device_id,
                          'device_owner': DEVICE_OWNER_COMPUTE}
                if current_port_status:
                    params['status'] = current_port_status
                port = port_obj.Port(self.ctx, **params)
                self._record_port_status_changed_helper(current_port_status,
                                                        previous_port_status,
                                                        port)

    def test_port_without_uuid_device_id_no_notify(self):
        port = port_obj.Port(self.ctx, id=self.port_uuid,
                             device_id='compute_probe:',
                             device_owner=DEVICE_OWNER_COMPUTE,
                             status=n_const.PORT_STATUS_ACTIVE)
        self._record_port_status_changed_helper(n_const.PORT_STATUS_ACTIVE,
                                                sql_attr.NO_VALUE,
                                                port)

    def test_port_without_device_owner_no_notify(self):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        port = port_obj.Port(self.ctx, id=self.port_uuid, device_id=device_id,
                             device_owner="",
                             status=n_const.PORT_STATUS_ACTIVE)
        self._record_port_status_changed_helper(n_const.PORT_STATUS_ACTIVE,
                                                sql_attr.NO_VALUE,
                                                port)

    def test_port_without_device_id_no_notify(self):
        port = port_obj.Port(self.ctx, id=self.port_uuid, device_id="",
                             device_owner=n_const.DEVICE_OWNER_DHCP,
                             status=n_const.PORT_STATUS_ACTIVE)
        self._record_port_status_changed_helper(n_const.PORT_STATUS_ACTIVE,
                                                sql_attr.NO_VALUE,
                                                port)

    def test_non_compute_instances_no_notify(self):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        port = port_obj.Port(self.ctx, id=self.port_uuid, device_id=device_id,
                             device_owner=n_const.DEVICE_OWNER_DHCP,
                             status=n_const.PORT_STATUS_ACTIVE)
        self._record_port_status_changed_helper(n_const.PORT_STATUS_ACTIVE,
                                                sql_attr.NO_VALUE,
                                                port)

    def _record_port_status_changed_helper(self, current_port_status,
                                           previous_port_status, port):

        if not (port.device_id and port.id and port.device_owner and
                port.device_owner.startswith(
                    n_const.DEVICE_OWNER_COMPUTE_PREFIX) and
                uuidutils.is_uuid_like(port.device_id)):
            return

        if (previous_port_status == n_const.PORT_STATUS_ACTIVE and
                current_port_status == n_const.PORT_STATUS_DOWN):
            event_name = nova.VIF_UNPLUGGED

        elif (previous_port_status in [sql_attr.NO_VALUE,
                                       n_const.PORT_STATUS_DOWN,
                                       n_const.PORT_STATUS_BUILD] and
              current_port_status in [n_const.PORT_STATUS_ACTIVE,
                                      n_const.PORT_STATUS_ERROR]):
            event_name = nova.VIF_PLUGGED

        else:
            return

        status = nova.NEUTRON_NOVA_EVENT_STATUS_MAP.get(current_port_status)
        self.nova_notifier.record_port_status_changed(port,
                                                      current_port_status,
                                                      previous_port_status,
                                                      None)

        event = {'server_uuid': port.device_id, 'status': status,
                 'name': event_name, 'tag': self.port_uuid}
        self.assertEqual(event, port._notify_event)

    def test_update_fixed_ip_changed(self):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        returned_obj = {'port':
                        {'device_owner': DEVICE_OWNER_COMPUTE,
                         'id': 'bee50827-bcee-4cc8-91c1-a27b0ce54222',
                         'device_id': device_id}}

        expected_event = {'server_uuid': device_id,
                          'name': 'network-changed',
                          'tag': returned_obj['port']['id']}
        event = self.nova_notifier.create_port_changed_event('update_port',
                                                             {}, returned_obj)
        self.assertEqual(event, expected_event)

    def test_create_floatingip_notify(self):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        returned_obj = {'floatingip':
                        {'port_id': 'bee50827-bcee-4cc8-91c1-a27b0ce54222'}}

        expected_event = {'server_uuid': device_id,
                          'name': 'network-changed',
                          'tag': returned_obj['floatingip']['port_id']}
        event = self.nova_notifier.create_port_changed_event(
            'create_floatingip', {}, returned_obj)
        self.assertEqual(event, expected_event)

    def test_create_floatingip_no_port_id_no_notify(self):
        returned_obj = {'floatingip':
                        {'port_id': None}}

        event = self.nova_notifier.create_port_changed_event(
            'create_floatingip', {}, returned_obj)
        self.assertIsNone(event)

    def test_delete_floatingip_notify(self):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        returned_obj = {'floatingip':
                        {'port_id': 'bee50827-bcee-4cc8-91c1-a27b0ce54222'}}

        expected_event = {'server_uuid': device_id,
                          'name': 'network-changed',
                          'tag': returned_obj['floatingip']['port_id']}
        event = self.nova_notifier.create_port_changed_event(
            'delete_floatingip', {}, returned_obj)
        self.assertEqual(expected_event, event)

    def test_delete_floatingip_deleted_port_no_notify(self):
        port_id = 'bee50827-bcee-4cc8-91c1-a27b0ce54222'
        with mock.patch.object(directory.get_plugin(), 'get_port',
                side_effect=n_exc.PortNotFound(port_id=port_id)):
            returned_obj = {'floatingip':
                            {'port_id': port_id}}
            event = self.nova_notifier.create_port_changed_event(
                'delete_floatingip', {}, returned_obj)
            self.assertIsNone(event)

    def test_delete_floatingip_no_port_id_no_notify(self):
        returned_obj = {'floatingip':
                        {'port_id': None}}

        event = self.nova_notifier.create_port_changed_event(
            'delete_floatingip', {}, returned_obj)
        self.assertIsNone(event)

    def test_associate_floatingip_notify(self):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        returned_obj = {'floatingip':
                        {'port_id': '5a39def4-3d3f-473d-9ff4-8e90064b9cc1'}}
        original_obj = {'port_id': None}

        expected_event = {'server_uuid': device_id,
                          'name': 'network-changed',
                          'tag': returned_obj['floatingip']['port_id']}
        event = self.nova_notifier.create_port_changed_event(
            'update_floatingip', original_obj, returned_obj)
        self.assertEqual(expected_event, event)

    def test_disassociate_floatingip_notify(self):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        returned_obj = {'floatingip': {'port_id': None}}
        original_obj = {'port_id': '5a39def4-3d3f-473d-9ff4-8e90064b9cc1'}

        expected_event = {'server_uuid': device_id,
                          'name': 'network-changed',
                          'tag': original_obj['port_id']}

        event = self.nova_notifier.create_port_changed_event(
            'update_floatingip', original_obj, returned_obj)
        self.assertEqual(expected_event, event)

    def test_no_notification_notify_nova_on_port_data_changes_false(self):
        cfg.CONF.set_override('notify_nova_on_port_data_changes', False)

        with mock.patch.object(self.nova_notifier,
                               'send_events') as send_events:
            self.nova_notifier.send_network_change('update_floatingip',
                                                   {}, {})
            self.assertFalse(send_events.called)

    @mock.patch('novaclient.client.Client')
    def test_nova_send_events_returns_bad_list(self, mock_client):
        create = mock_client().server_external_events.create
        create.return_value = 'i am a string!'
        self.nova_notifier.send_events([])
        create.assert_called()

    @mock.patch('novaclient.client.Client')
    def test_nova_send_event_rasies_404(self, mock_client):
        create = mock_client().server_external_events.create
        create.return_value = nova_exceptions.NotFound
        self.nova_notifier.send_events([])
        create.assert_called()

    @mock.patch('novaclient.client.Client')
    def test_nova_send_events_raises_connect_exc(self, mock_client):
        create = mock_client().server_external_events.create
        create.side_effect = (
            ks_exc.ConnectFailure, ks_exc.ConnectTimeout, [])
        self.nova_notifier.send_events([])
        self.assertEqual(3, create.call_count)

    @mock.patch('novaclient.client.Client')
    def test_nova_send_events_raises(self, mock_client):
        create = mock_client().server_external_events.create
        create.return_value = Exception
        self.nova_notifier.send_events([])
        create.assert_called()

    @mock.patch('novaclient.client.Client')
    def test_nova_send_events_returns_non_200(self, mock_client):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        create = mock_client().server_external_events.create
        create.return_value = [
            {'code': 404,
             'name': 'network-changed',
             'server_uuid': device_id}]
        self.nova_notifier.send_events([{'name': 'network-changed',
                                         'server_uuid': device_id}])
        create.assert_called()

    @mock.patch('novaclient.client.Client')
    def test_nova_send_events_return_200(self, mock_client):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        create = mock_client().server_external_events.create
        create.return_value = [
            {'code': 200,
             'name': 'network-changed',
             'server_uuid': device_id}]
        self.nova_notifier.send_events([{'name': 'network-changed',
                                         'server_uuid': device_id}])
        create.assert_called()

    @mock.patch('novaclient.client.Client')
    def test_nova_send_events_multiple(self, mock_client):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        create = mock_client().server_external_events.create
        create.return_value = [
            {'code': 200,
             'name': 'network-changed',
             'server_uuid': device_id},
            {'code': 200,
             'name': 'network-changed',
             'server_uuid': device_id}]
        self.nova_notifier.send_events([
            {'name': 'network-changed', 'server_uuid': device_id},
            {'name': 'network-changed', 'server_uuid': device_id}])
        create.assert_called()

    def test_reassociate_floatingip_without_disassociate_event(self):
        returned_obj = {'floatingip':
                        {'port_id': 'f5348a16-609a-4971-b0f0-4b8def5235fb'}}
        original_obj = {'port_id': '5a39def4-3d3f-473d-9ff4-8e90064b9cc1'}
        self.nova_notifier._waiting_to_send = True
        self.nova_notifier.send_network_change(
            'update_floatingip', original_obj, returned_obj)
        self.assertEqual(
            2, len(self.nova_notifier.batch_notifier._pending_events.queue))

        returned_obj_non = {'floatingip': {'port_id': None}}
        event_dis = self.nova_notifier.create_port_changed_event(
            'update_floatingip', original_obj, returned_obj_non)
        event_assoc = self.nova_notifier.create_port_changed_event(
            'update_floatingip', original_obj, returned_obj)
        self.assertEqual(
            self.nova_notifier.batch_notifier._pending_events.get(), event_dis)
        self.assertEqual(
            self.nova_notifier.batch_notifier._pending_events.get(),
            event_assoc)

    def test_delete_port_notify(self):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        port_id = 'bee50827-bcee-4cc8-91c1-a27b0ce54222'
        returned_obj = {'port':
                        {'device_owner': DEVICE_OWNER_COMPUTE,
                         'id': port_id,
                         'device_id': device_id}}

        expected_event = {'server_uuid': device_id,
                          'name': nova.VIF_DELETED,
                          'tag': port_id}
        event = self.nova_notifier.create_port_changed_event('delete_port',
                                                             {}, returned_obj)
        self.assertEqual(expected_event, event)

    @mock.patch('novaclient.client.Client')
    def test_endpoint_types(self, mock_client):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        batched_events = [{'code': 200,
                           'name': 'network-changed',
                           'server_uuid': device_id}]
        response = [{'name': 'network-changed',
                     'server_uuid': device_id}]
        mock_client.server_external_events.create.return_value = (
            batched_events)
        self.nova_notifier.send_events(response)
        mock_client.assert_called_once_with(
            api_versions.APIVersion(nova.NOVA_API_VERSION),
            connect_retries=3,
            session=mock.ANY,
            region_name=cfg.CONF.nova.region_name,
            endpoint_type='public',
            extensions=mock.ANY,
            global_request_id=mock.ANY)

        mock_client.reset_mock()
        cfg.CONF.set_override('endpoint_type', 'internal', 'nova')
        mock_client.server_external_events.create.return_value = (
            batched_events)
        self.nova_notifier.send_events(response)
        mock_client.assert_called_once_with(
            api_versions.APIVersion(nova.NOVA_API_VERSION),
            connect_retries=3,
            session=mock.ANY,
            region_name=cfg.CONF.nova.region_name,
            endpoint_type='internal',
            extensions=mock.ANY,
            global_request_id=mock.ANY)

    def test_notify_port_active_direct(self):
        device_id = '32102d7b-1cf4-404d-b50a-97aae1f55f87'
        port_id = 'bee50827-bcee-4cc8-91c1-a27b0ce54222'
        port = port_obj.Port(self.ctx, id=port_id, device_id=device_id,
                             device_owner=DEVICE_OWNER_COMPUTE)
        expected_event = {'server_uuid': device_id,
                          'name': nova.VIF_PLUGGED,
                          'status': 'completed',
                          'tag': port_id}
        self.nova_notifier.notify_port_active_direct(port)

        self.assertEqual(
            1, len(self.nova_notifier.batch_notifier._pending_events.queue))
        self.assertEqual(
            expected_event,
            self.nova_notifier.batch_notifier._pending_events.get())

    def test_notify_concurrent_enable_flag_update(self):
        # This test assumes Neutron server uses eventlet.
        # NOTE(ralonsoh): the exceptions raise inside a thread won't stop the
        # test. The checks are stored in "_queue" and tested at the end of the
        # test execution.
        _queue = eventlet.queue.Queue()

        def _local_executor(thread_idx):
            # This thread has not yet initialized the local "enable" flag.
            _queue.put(getattr(nova._notifier_store, 'enable', None) is None)
            eventlet.sleep(0)  # Next thread execution.
            new_enable = bool(thread_idx % 2)
            with self.nova_notifier.context_enabled(new_enable):
                # At this point, the Nova Notifier should have updated the
                # "enable" flag.
                _queue.put(new_enable == nova._notifier_store.enable)
                eventlet.sleep(0)  # Next thread execution.
                _queue.put(new_enable == nova._notifier_store.enable)
            _queue.put(nova.NOTIFIER_ENABLE_DEFAULT ==
                       nova._notifier_store.enable)

        num_threads = 20
        pool = eventlet.GreenPool(num_threads)
        for idx in range(num_threads):
            pool.spawn(_local_executor, idx)
        pool.waitall()
        try:
            while True:
                self.assertTrue(_queue.get(block=False))
        except queue.Empty:
            pass



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import address_group
from neutron_lib.api.definitions import address_scope
from neutron_lib.api.definitions import agent as agent_def
from neutron_lib.api.definitions import allowedaddresspairs
from neutron_lib.api.definitions import auto_allocated_topology
from neutron_lib.api.definitions import availability_zone as az_def
from neutron_lib.api.definitions import bgp
from neutron_lib.api.definitions import bgp_4byte_asn
from neutron_lib.api.definitions import bgp_dragentscheduler
from neutron_lib.api.definitions import default_subnetpools
from neutron_lib.api.definitions import dhcpagentscheduler
from neutron_lib.api.definitions import dns
from neutron_lib.api.definitions import dns_domain_keywords
from neutron_lib.api.definitions import dns_domain_ports
from neutron_lib.api.definitions import expose_port_forwarding_in_fip
from neutron_lib.api.definitions import external_net
from neutron_lib.api.definitions import extra_dhcp_opt
from neutron_lib.api.definitions import extraroute
from neutron_lib.api.definitions import filter_validation
from neutron_lib.api.definitions import fip_pf_description
from neutron_lib.api.definitions import fip_pf_detail
from neutron_lib.api.definitions import fip_pf_port_range
from neutron_lib.api.definitions import fip_port_details
from neutron_lib.api.definitions import floating_ip_port_forwarding
from neutron_lib.api.definitions import floatingip_pools
from neutron_lib.api.definitions import l3
from neutron_lib.api.definitions import l3_ext_gw_mode
from neutron_lib.api.definitions import logging
from neutron_lib.api.definitions import multiprovidernet
from neutron_lib.api.definitions import network_availability_zone
from neutron_lib.api.definitions import network_ha
from neutron_lib.api.definitions import network_ip_availability
from neutron_lib.api.definitions import network_mtu
from neutron_lib.api.definitions import network_mtu_writable
from neutron_lib.api.definitions import pagination
from neutron_lib.api.definitions import port_device_profile
from neutron_lib.api.definitions import port_hardware_offload_type
from neutron_lib.api.definitions import port_mac_address_regenerate
from neutron_lib.api.definitions import port_numa_affinity_policy
from neutron_lib.api.definitions import port_resource_request
from neutron_lib.api.definitions import port_security
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import portbindings_extended as pbe_ext
from neutron_lib.api.definitions import project_id
from neutron_lib.api.definitions import provider_net
from neutron_lib.api.definitions import qos
from neutron_lib.api.definitions import qos_bw_limit_direction
from neutron_lib.api.definitions import qos_default
from neutron_lib.api.definitions import qos_gateway_ip
from neutron_lib.api.definitions import qos_rule_type_details
from neutron_lib.api.definitions import qos_rule_type_filter
from neutron_lib.api.definitions import qos_rules_alias
from neutron_lib.api.definitions import quota_check_limit
from neutron_lib.api.definitions import rbac_address_scope
from neutron_lib.api.definitions import rbac_security_groups
from neutron_lib.api.definitions import router_availability_zone as raz_def
from neutron_lib.api.definitions import security_groups_normalized_cidr
from neutron_lib.api.definitions import security_groups_remote_address_group
from neutron_lib.api.definitions import \
    security_groups_rules_belongs_to_default_sg
from neutron_lib.api.definitions import security_groups_shared_filtering
from neutron_lib.api.definitions import segment as seg_def
from neutron_lib.api.definitions import sorting
from neutron_lib.api.definitions import stateful_security_group
from neutron_lib.api.definitions import subnet_dns_publish_fixed_ip
from neutron_lib.api.definitions import subnet_service_types
from neutron_lib.api.definitions import trunk
from neutron_lib.api.definitions import uplink_status_propagation
from neutron_lib.api.definitions import vlantransparent
from neutron_lib.api.definitions import vpn
from neutron_lib.api.definitions import vpn_endpoint_groups
from neutron_lib import constants

from neutron.extensions import quotasv2_detail
from neutron.extensions import security_groups_default_rules

# NOTE(russellb) This remains in its own file (vs constants.py) because we want
# to be able to easily import it and export the info without any dependencies
# on external imports.

# NOTE(russellb) If you update these lists, please also update
# doc/source/admin/ovn/features.rst and the current release note.
ML2_SUPPORTED_API_EXTENSIONS_OVN_L3 = [
    l3.ALIAS,
    extraroute.ALIAS,
    l3_ext_gw_mode.ALIAS,
    fip_pf_detail.ALIAS,
    fip_port_details.ALIAS,
    floatingip_pools.ALIAS,
    pagination.ALIAS,
    'qos-fip',
    qos_gateway_ip.ALIAS,
    sorting.ALIAS,
    project_id.ALIAS,
    dns.ALIAS,
    dns_domain_keywords.ALIAS,
    dns_domain_ports.ALIAS,
    subnet_dns_publish_fixed_ip.ALIAS,
    agent_def.ALIAS,
    az_def.ALIAS,
    raz_def.ALIAS,
]
ML2_SUPPORTED_API_EXTENSIONS = [
    address_group.ALIAS,
    address_scope.ALIAS,
    agent_def.ALIAS,
    allowedaddresspairs.ALIAS,
    auto_allocated_topology.ALIAS,
    az_def.ALIAS,
    portbindings.ALIAS,
    pbe_ext.ALIAS,
    default_subnetpools.ALIAS,
    dhcpagentscheduler.ALIAS,
    dns.ALIAS,
    external_net.ALIAS,
    extra_dhcp_opt.ALIAS,
    filter_validation.ALIAS,
    multiprovidernet.ALIAS,
    network_ha.ALIAS,
    network_mtu.ALIAS,
    network_mtu_writable.ALIAS,
    network_availability_zone.ALIAS,
    network_ip_availability.ALIAS,
    port_device_profile.ALIAS,
    port_hardware_offload_type.ALIAS,
    port_mac_address_regenerate.ALIAS,
    port_numa_affinity_policy.ALIAS,
    port_security.ALIAS,
    provider_net.ALIAS,
    port_resource_request.ALIAS,
    qos.ALIAS,
    qos_bw_limit_direction.ALIAS,
    qos_default.ALIAS,
    qos_rule_type_details.ALIAS,
    qos_rule_type_filter.ALIAS,
    qos_rules_alias.ALIAS,
    'quotas',
    quota_check_limit.ALIAS,
    quotasv2_detail.ALIAS,
    rbac_address_scope.ALIAS,
    'rbac-policies',
    rbac_security_groups.ALIAS,
    'standard-attr-revisions',
    'security-group',
    security_groups_default_rules.ALIAS,
    security_groups_normalized_cidr.ALIAS,
    security_groups_remote_address_group.ALIAS,
    security_groups_rules_belongs_to_default_sg.ALIAS,
    security_groups_shared_filtering.ALIAS,
    stateful_security_group.ALIAS,
    'standard-attr-description',
    constants.SUBNET_ALLOCATION_EXT_ALIAS,
    'standard-attr-tag',
    'standard-attr-timestamp',
    subnet_service_types.ALIAS,
    trunk.ALIAS,
    seg_def.ALIAS,
    expose_port_forwarding_in_fip.ALIAS,
    fip_pf_description.ALIAS,
    fip_pf_port_range.ALIAS,
    floating_ip_port_forwarding.ALIAS,
    vlantransparent.ALIAS,
    logging.ALIAS,
    vpn.ALIAS,
    vpn_endpoint_groups.ALIAS,
    bgp.ALIAS,
    bgp_4byte_asn.ALIAS,
    bgp_dragentscheduler.ALIAS,
    uplink_status_propagation.ALIAS,
]



# Copyright 2023 Canonical Ltd.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import l3_enable_default_route_ecmp as apidef
from neutron_lib.api import extensions


class L3_enable_default_route_ecmp(extensions.APIExtensionDescriptor):

    api_definition = apidef

    def __init__(self):
        # NOTE(fnordahl): Temporary fix awaiting permanent fix in neutron-lib,
        # drop when change I9096685fb79a84e11a8547a5aaa16f7f2df48a56 is merged.
        apidef.RESOURCE_ATTRIBUTE_MAP[
            apidef.COLLECTION_NAME][apidef.ENABLE_DEFAULT_ROUTE_ECMP].update(
                {'default': None})
        super().__init__()



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from datetime import datetime
import errno
import os
import shutil
from unittest import mock
import warnings

import fixtures
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils
from oslo_utils import uuidutils

from neutron.agent.linux import utils
from neutron.api import extensions as exts
from neutron.common import utils as n_utils
from neutron.conf.agent import common as config
from neutron.conf.agent import ovs_conf
from neutron.conf.plugins.ml2 import config as ml2_config
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
# Load all the models to register them into SQLAlchemy metadata before using
# the SqlFixture
from neutron.db import models  # noqa
from neutron import manager
from neutron.plugins.ml2.drivers.ovn.agent import neutron_agent
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb.extensions import \
    placement as ovn_client_placement
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import worker
from neutron.plugins.ml2.drivers import type_geneve  # noqa
from neutron import service  # noqa
from neutron.tests import base
from neutron.tests.common import base as common_base
from neutron.tests.common import helpers
from neutron.tests.functional.resources import process
from neutron.tests.unit.extensions import test_securitygroup
from neutron.tests.unit.plugins.ml2 import test_plugin
import neutron.wsgi

LOG = log.getLogger(__name__)

# This is the directory from which infra fetches log files for functional tests
DEFAULT_LOG_DIR = os.path.join(helpers.get_test_log_path(),
                               'dsvm-functional-logs')


def config_decorator(method_to_decorate, config_tuples):
    def wrapper(*args, **kwargs):
        method_to_decorate(*args, **kwargs)
        for config_tuple in config_tuples:
            cfg.CONF.set_override(*config_tuple)
    return wrapper


class BaseLoggingTestCase(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        # NOTE(slaweq): Because of issue with stestr and Python3, we need
        # to avoid too much output to be produced during tests, so we will
        # ignore python warnings here
        warnings.simplefilter("ignore")
        base.setup_test_logging(
            cfg.CONF, DEFAULT_LOG_DIR, "%s.txt" % self.id())
        cfg.CONF.set_override('use_helper_for_ns_read', False, group='AGENT')


class BaseSudoTestCase(BaseLoggingTestCase):
    """Base class for tests requiring invocation of commands via a root helper.

    This class skips (during setUp) its tests unless sudo is enabled, ie:
    OS_SUDO_TESTING is set to '1' or 'True' in the test execution environment.
    This is intended to allow developers to run the functional suite (e.g. tox
    -e functional) without test failures if sudo invocations are not allowed.

    Running sudo tests in the upstream gate jobs
    (*-neutron-dsvm-functional) requires the additional step of
    setting OS_ROOTWRAP_CMD to the rootwrap command configured by
    devstack, e.g.

      sudo /usr/local/bin/neutron-rootwrap /etc/neutron/rootwrap.conf

    Gate jobs do not allow invocations of sudo without rootwrap to
    ensure that rootwrap configuration gets as much testing as
    possible.
    """

    def setUp(self):
        super(BaseSudoTestCase, self).setUp()
        if not base.bool_from_env('OS_SUDO_TESTING'):
            self.skipTest('Testing with sudo is not enabled')
        self.setup_rootwrap()
        config.setup_privsep()
        self._override_default_config()

    @common_base.no_skip_on_missing_deps
    def check_command(self, cmd, error_text, skip_msg, run_as_root=False):
        try:
            utils.execute(cmd, run_as_root=run_as_root)
        except RuntimeError as e:
            if error_text in str(e):
                self.skipTest(skip_msg)
            raise

    @staticmethod
    def _override_default_config():
        # NOTE(ralonsoh): once https://review.opendev.org/#/c/641681/ is
        # merged, we should increase the default value of those new parameters.
        ovs_agent_opts = [('ovsdb_timeout', 30, 'OVS'),
                          ('ovsdb_debug', True, 'OVS'),
                          ]
        ovs_agent_decorator = config_decorator(
            ovs_conf.register_ovs_agent_opts, ovs_agent_opts)
        mock.patch.object(ovs_conf, 'register_ovs_agent_opts',
                          new=ovs_agent_decorator).start()


class TestOVNFunctionalBase(test_plugin.Ml2PluginV2TestCase,
                            BaseLoggingTestCase):

    OVS_DISTRIBUTION = 'openvswitch'
    OVN_DISTRIBUTION = 'ovn'
    OVN_SCHEMA_FILES = ['ovn-nb.ovsschema', 'ovn-sb.ovsschema']

    _mechanism_drivers = ['logger', 'ovn']
    _extension_drivers = ['port_security']
    _counter = 0
    l3_plugin = 'neutron.services.ovn_l3.plugin.OVNL3RouterPlugin'

    def preSetUp(self):
        self.temp_dir = self.useFixture(fixtures.TempDir()).path
        self._start_ovsdb_server()

    def setUp(self, maintenance_worker=False, service_plugins=None):
        ml2_config.cfg.CONF.set_override('extension_drivers',
                                     self._extension_drivers,
                                     group='ml2')
        ml2_config.cfg.CONF.set_override('tenant_network_types',
                                     ['geneve'],
                                     group='ml2')
        ml2_config.cfg.CONF.set_override('vni_ranges',
                                     ['1:65536'],
                                     group='ml2_type_geneve')
        # ensure viable minimum is set for OVN's Geneve
        ml2_config.cfg.CONF.set_override('max_header_size', 38,
                                         group='ml2_type_geneve')
        ovn_conf.register_opts()
        ovn_conf.cfg.CONF.set_override('dns_servers',
                                       ['10.10.10.10'],
                                       group='ovn')
        ovn_conf.cfg.CONF.set_override('api_workers', 1)

        self.addCleanup(exts.PluginAwareExtensionManager.clear_instance)
        self.ovsdb_server_mgr = None
        self._service_plugins = service_plugins
        super(TestOVNFunctionalBase, self).setUp()
        self.test_log_dir = os.path.join(DEFAULT_LOG_DIR, self.id())
        base.setup_test_logging(
            cfg.CONF, self.test_log_dir, "testrun.txt")

        mm = directory.get_plugin().mechanism_manager
        self.mech_driver = mm.mech_drivers['ovn'].obj
        self.l3_plugin = directory.get_plugin(constants.L3)
        self.segments_plugin = directory.get_plugin('segments')
        # OVN does not use RPC: disable it for port-forwarding tests
        self.pf_plugin = manager.NeutronManager.load_class_for_provider(
            'neutron.service_plugins', 'port_forwarding')()
        self.pf_plugin._rpc_notifications_required = False
        self.log_plugin = directory.get_plugin(constants.LOG_API)
        if not self.log_plugin:
            self.log_plugin = manager.NeutronManager.load_class_for_provider(
                'neutron.service_plugins', 'log')()
            directory.add_plugin(constants.LOG_API, self.log_plugin)
            self.log_plugin.driver_manager.register_driver(
                self.mech_driver.log_driver)
        self.mech_driver.log_driver.plugin_driver = self.mech_driver
        self.mech_driver.log_driver._log_plugin_property = None
        self.ovn_northd_mgr = None
        self.maintenance_worker = maintenance_worker
        mock.patch(
            'neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb.'
            'maintenance.MaintenanceThread').start()
        mock.patch(
            'neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb.'
            'maintenance.HashRingHealthCheckPeriodics').start()
        self._start_idls()
        self._start_ovn_northd()
        self.addCleanup(self._reset_agent_cache_singleton)
        self.addCleanup(self._reset_ovn_client_placement_extension)
        plugin = directory.get_plugin()
        mock.patch.object(
            plugin, 'get_default_security_group_rules',
            return_value=copy.deepcopy(
                test_securitygroup.RULES_TEMPLATE_FOR_DEFAULT_SG)).start()

    def _reset_agent_cache_singleton(self):
        neutron_agent.AgentCache._instance = None

    def _reset_ovn_client_placement_extension(self):
        ovn_client_placement.OVNClientPlacementExtension._instance = None

    def _get_install_share_path(self):
        lookup_paths = set()
        for installation in ['local', '']:
            for distribution in [self.OVN_DISTRIBUTION, self.OVS_DISTRIBUTION]:
                exists = True
                for ovn_file in self.OVN_SCHEMA_FILES:
                    path = os.path.join(os.path.sep, 'usr', installation,
                                        'share', distribution, ovn_file)
                    exists &= os.path.isfile(path)
                    lookup_paths.add(os.path.dirname(path))
                if exists:
                    return os.path.dirname(path)
        msg = 'Either ovn-nb.ovsschema and/or ovn-sb.ovsschema are missing. '
        msg += 'Looked for schemas in paths:' + ', '.join(sorted(lookup_paths))
        raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), msg)

    def get_additional_service_plugins(self):
        p = super(TestOVNFunctionalBase, self).get_additional_service_plugins()
        p.update({'revision_plugin_name': 'revisions',
                  'segments': 'neutron.services.segments.plugin.Plugin'})
        if self._service_plugins:
            p.update(self._service_plugins)
        return p

    @property
    def _ovsdb_protocol(self):
        return self.get_ovsdb_server_protocol()

    def get_ovsdb_server_protocol(self):
        return 'unix'

    def _start_ovn_northd(self):
        if not self.ovsdb_server_mgr:
            return

        def wait_for_northd():
            try:
                self.nb_api.nb_global
            except StopIteration:
                LOG.debug("NB_Global is not ready yet")
                return False

            try:
                next(iter(self.sb_api.db_list_rows('SB_Global').execute(
                    check_error=True)))
            except StopIteration:
                LOG.debug("SB_Global is not ready yet")
                return False
            except KeyError:
                # Maintenance worker doesn't register SB_Global therefore
                # we don't need to wait for it
                LOG.debug("SB_Global is not registered in this IDL")

            return True

        timeout = 20
        ovn_nb_db = self.ovsdb_server_mgr.get_ovsdb_connection_path('nb')
        ovn_sb_db = self.ovsdb_server_mgr.get_ovsdb_connection_path('sb')
        LOG.debug("Starting OVN northd")
        self.ovn_northd_mgr = self.useFixture(
            process.OvnNorthd(self.temp_dir,
                              ovn_nb_db, ovn_sb_db,
                              protocol=self._ovsdb_protocol))
        LOG.debug("OVN northd started: %r", self.ovn_northd_mgr)
        n_utils.wait_until_true(
            wait_for_northd, timeout, sleep=1,
            exception=Exception(
                "ovn-northd didn't initialize OVN DBs in %d"
                "seconds" % timeout))

    def _start_ovsdb_server(self):
        # Start 2 ovsdb-servers one each for OVN NB DB and OVN SB DB
        # ovsdb-server with OVN SB DB can be used to test the chassis up/down
        # events.
        install_share_path = self._get_install_share_path()
        self.ovsdb_server_mgr = self.useFixture(
            process.OvsdbServer(self.temp_dir, install_share_path,
                                ovn_nb_db=True, ovn_sb_db=True,
                                protocol=self._ovsdb_protocol))
        LOG.debug("OVSDB server manager instantiated: %r",
                  self.ovsdb_server_mgr)
        set_cfg = cfg.CONF.set_override
        set_cfg('ovn_nb_connection',
                self.ovsdb_server_mgr.get_ovsdb_connection_path(), 'ovn')
        set_cfg('ovn_sb_connection',
                self.ovsdb_server_mgr.get_ovsdb_connection_path(
                    db_type='sb'), 'ovn')
        set_cfg('ovn_nb_private_key', self.ovsdb_server_mgr.private_key, 'ovn')
        set_cfg('ovn_nb_certificate', self.ovsdb_server_mgr.certificate, 'ovn')
        set_cfg('ovn_nb_ca_cert', self.ovsdb_server_mgr.ca_cert, 'ovn')
        set_cfg('ovn_sb_private_key', self.ovsdb_server_mgr.private_key, 'ovn')
        set_cfg('ovn_sb_certificate', self.ovsdb_server_mgr.certificate, 'ovn')
        set_cfg('ovn_sb_ca_cert', self.ovsdb_server_mgr.ca_cert, 'ovn')

        # NOTE(mjozefcz): We can find occasional functional test
        # failures because of low timeout value - set it to 30
        # seconds, should be enough. More info: 1868110
        cfg.CONF.set_override(
            'ovsdb_connection_timeout', 30,
            'ovn')
        self.addCleanup(self._collect_processes_logs)

    def _start_idls(self):
        class TriggerCls(mock.MagicMock):
            def trigger(self):
                pass

        trigger_cls = TriggerCls()
        if self.maintenance_worker:
            trigger_cls.trigger.__self__.__class__ = worker.MaintenanceWorker
            cfg.CONF.set_override('neutron_sync_mode', 'off', 'ovn')
        else:
            trigger_cls.trigger.__self__.__class__ = neutron.wsgi.WorkerService

        self.addCleanup(self.stop)
        # NOTE(ralonsoh): do not access to the DB at exit when the SQL
        # connection is already closed, to avoid useless exception messages.
        mock.patch.object(
            self.mech_driver, '_remove_node_from_hash_ring').start()
        self.mech_driver.pre_fork_initialize(
            mock.ANY, mock.ANY, trigger_cls.trigger)

        # mech_driver.post_fork_initialize creates the IDL connections
        self.mech_driver.post_fork_initialize(
            mock.ANY, mock.ANY, trigger_cls.trigger)

        self.nb_api = self.mech_driver.nb_ovn
        self.sb_api = self.mech_driver.sb_ovn

    def _collect_processes_logs(self):
        timestamp = datetime.now().strftime('%y-%m-%d_%H-%M-%S')
        for database in ("nb", "sb"):
            for file_suffix in ("log", "db"):
                src_filename = "ovn_%(db)s.%(suffix)s" % {
                    'db': database,
                    'suffix': file_suffix
                }
                dst_filename = "ovn_%(db)s-%(timestamp)s.%(suffix)s" % {
                    'db': database,
                    'suffix': file_suffix,
                    'timestamp': timestamp,
                }
                self._copy_log_file(src_filename, dst_filename)

        # Copy northd logs
        northd_log = "ovn_northd"
        dst_northd = "%(northd)s-%(timestamp)s.log" % {
            "northd": northd_log,
            "timestamp": timestamp,
        }
        self._copy_log_file("%s.log" % northd_log, dst_northd)

    def _copy_log_file(self, src_filename, dst_filename):
        """Copy log file from temporary dict to the test directory."""
        filepath = os.path.join(self.temp_dir, src_filename)
        shutil.copyfile(
            filepath, os.path.join(self.test_log_dir, dst_filename))

    def stop(self):
        if self.maintenance_worker:
            self.mech_driver.nb_synchronizer.stop()
            self.mech_driver.sb_synchronizer.stop()
        self.mech_driver.nb_ovn.ovsdb_connection.stop()
        self.mech_driver.sb_ovn.ovsdb_connection.stop()

    def restart(self):
        self.stop()

        if self.ovsdb_server_mgr:
            self.ovsdb_server_mgr.stop()
        if self.ovn_northd_mgr:
            self.ovn_northd_mgr.stop()

        self.ovsdb_server_mgr.delete_dbs()
        self._start_ovsdb_server()
        self._start_idls()
        self._start_ovn_northd()

    def add_fake_chassis(self, host, physical_nets=None, external_ids=None,
                         name=None, azs=None, enable_chassis_as_gw=False,
                         enable_chassis_as_extport=False, other_config=None):
        def append_cms_options(ext_ids, value):
            if 'ovn-cms-options' not in ext_ids:
                ext_ids['ovn-cms-options'] = value
            else:
                ext_ids['ovn-cms-options'] += ',' + value

        physical_nets = physical_nets or []
        external_ids = external_ids or {}
        other_config = other_config or {}
        if azs is None:
            azs = ['ovn']
        if azs:
            append_cms_options(other_config, 'availability-zones=')
            other_config['ovn-cms-options'] += ':'.join(azs)
        if enable_chassis_as_gw:
            append_cms_options(other_config, 'enable-chassis-as-gw')
        if enable_chassis_as_extport:
            append_cms_options(other_config, 'enable-chassis-as-extport-host')

        bridge_mapping = ",".join(["%s:br-provider%s" % (phys_net, i)
                                  for i, phys_net in enumerate(physical_nets)])
        if name is None:
            name = uuidutils.generate_uuid()
        other_config['ovn-bridge-mappings'] = bridge_mapping
        # We'll be using different IP addresses every time for the Encap of
        # the fake chassis as the SB schema doesn't allow to have two entries
        # with same (ip,type) pairs as of OVS 2.11. This shouldn't have any
        # impact as the tunnels won't get created anyways since ovn-controller
        # is not running. Ideally we shouldn't be creating more than 255
        # fake chassis but from the SB db point of view, 'ip' column can be
        # any string so we could add entries with ip='172.24.4.1000'.
        self._counter += 1
        chassis = self.sb_api.chassis_add(
            name, ['geneve'], '172.24.4.%d' % self._counter,
            external_ids=external_ids, hostname=host,
            other_config=other_config).execute(check_error=True)
        nb_cfg_timestamp = timeutils.utcnow_ts() * 1000
        self.sb_api.db_create(
            'Chassis_Private', name=name, external_ids=external_ids,
            chassis=chassis.uuid, nb_cfg_timestamp=nb_cfg_timestamp
        ).execute(check_error=True)
        return name

    def del_fake_chassis(self, chassis, if_exists=True):
        self.sb_api.chassis_del(
            chassis, if_exists=if_exists).execute(check_error=True)
        self.sb_api.db_destroy(
            'Chassis_Private', chassis).execute(check_error=True)



# Copyright 2018 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import random
import threading
from unittest import mock

import netaddr
from neutron_lib import constants as n_cons
from oslo_utils import uuidutils
from pyroute2.ipdb import routes as ipdb_routes
from pyroute2.iproute import linux as iproute_linux
from pyroute2.netlink import exceptions as netlink_exc
from pyroute2.netlink import rtnl
import testtools

from neutron.agent.linux import ip_lib
from neutron.agent.linux import utils as linux_utils
from neutron.common import utils as common_utils
from neutron.privileged.agent.linux import ip_lib as priv_ip_lib
from neutron.tests.common import net_helpers
from neutron.tests.functional import base as functional_base


class GetDeviceNamesTestCase(functional_base.BaseSudoTestCase):

    def _remove_ns(self, namespace):
        priv_ip_lib.remove_netns(namespace)

    def test_get_device_names(self):
        namespace = 'ns_test-' + uuidutils.generate_uuid()
        priv_ip_lib.create_netns(namespace)
        self.addCleanup(self._remove_ns, namespace)
        interfaces = ['int_01', 'int_02', 'int_03', 'int_04', 'int_05']
        interfaces_to_check = (interfaces + ip_lib.FB_TUNNEL_DEVICE_NAMES +
                               [ip_lib.LOOPBACK_DEVNAME])
        for interface in interfaces:
            priv_ip_lib.create_interface(interface, namespace, 'dummy')

        device_names = priv_ip_lib.get_device_names(namespace)
        self.assertGreater(len(device_names), 0)
        for name in device_names:
            self.assertIn(name, interfaces_to_check)

        for interface in interfaces:
            priv_ip_lib.delete_interface(interface, namespace)

        device_names = priv_ip_lib.get_device_names(namespace)
        self.assertGreater(len(device_names), 0)
        for name in device_names:
            self.assertNotIn(name, interfaces)


class GetLinkDevicesTestCase(functional_base.BaseSudoTestCase):

    def setUp(self):
        super().setUp()
        self.namespace = 'ns_test-' + uuidutils.generate_uuid()
        priv_ip_lib.create_netns(self.namespace)
        self.addCleanup(self._remove_ns, self.namespace)
        self.interfaces = ['int_01', 'int_02']
        self.interfaces_to_exclude = (ip_lib.FB_TUNNEL_DEVICE_NAMES +
                                      [ip_lib.LOOPBACK_DEVNAME])

    def _remove_ns(self, namespace):
        priv_ip_lib.remove_netns(namespace)

    def test_get_link_devices_lo(self):
        devices = priv_ip_lib.get_link_devices(self.namespace)
        self.assertGreater(len(devices), 0)
        for device in devices:
            if linux_utils.get_attr(device, 'IFLA_IFNAME') != 'lo':
                continue
            self.assertIsNone(linux_utils.get_attr(device, 'IFLA_LINKINFO'))
            break
        else:
            self.fail('Device "lo" not found')

    def test_get_link_devices_dummy(self):
        interfaces_tested = []
        for interface in self.interfaces:
            priv_ip_lib.create_interface(interface, self.namespace, 'dummy')

        devices = priv_ip_lib.get_link_devices(self.namespace)
        self.assertGreater(len(devices), 0)
        for device in devices:
            name = linux_utils.get_attr(device, 'IFLA_IFNAME')
            if name in self.interfaces_to_exclude:
                continue
            self.assertIn(name, self.interfaces)
            ifla_linkinfo = linux_utils.get_attr(device, 'IFLA_LINKINFO')
            self.assertEqual(
                linux_utils.get_attr(ifla_linkinfo, 'IFLA_INFO_KIND'),
                'dummy')
            interfaces_tested.append(name)
        self.assertEqual(sorted(interfaces_tested), sorted(self.interfaces))

    def test_get_link_devices_vlan(self):
        interfaces_tested = []
        vlan_interfaces = []
        vlan_id = 1000
        for interface in self.interfaces:
            priv_ip_lib.create_interface(interface, self.namespace, 'dummy')
            vlan_interface = interface + '_' + str(vlan_id)
            vlan_interfaces.append(vlan_interface)
            priv_ip_lib.create_interface(
                vlan_interface, self.namespace, 'vlan',
                physical_interface=interface, vlan_id=vlan_id)
            vlan_id += 1

        devices = priv_ip_lib.get_link_devices(self.namespace)
        self.assertGreater(len(devices), 0)
        device_name_index = {}
        for device in devices:
            name = linux_utils.get_attr(device, 'IFLA_IFNAME')
            device_name_index[name] = device['index']

        for device in devices:
            name = linux_utils.get_attr(device, 'IFLA_IFNAME')
            if name in self.interfaces_to_exclude:
                continue
            self.assertIn(name, self.interfaces + vlan_interfaces)
            ifla_linkinfo = linux_utils.get_attr(device, 'IFLA_LINKINFO')
            if name in vlan_interfaces:
                self.assertEqual(
                    linux_utils.get_attr(ifla_linkinfo, 'IFLA_INFO_KIND'),
                    'vlan')
                ifla_infodata = linux_utils.get_attr(ifla_linkinfo,
                                                     'IFLA_INFO_DATA')
                vlan_id = int(name.split('_')[-1])
                self.assertEqual(
                    linux_utils.get_attr(ifla_infodata, 'IFLA_VLAN_ID'),
                    vlan_id)
                vlan_link_name = self.interfaces[vlan_interfaces.index(name)]
                vlan_link_index = device_name_index[vlan_link_name]
                self.assertEqual(vlan_link_index,
                                 linux_utils.get_attr(device, 'IFLA_LINK'))
            interfaces_tested.append(name)
        self.assertEqual(sorted(interfaces_tested),
                         sorted(self.interfaces + vlan_interfaces))

    def test_get_link_devices_vxlan(self):
        interfaces_tested = []
        vxlan_interfaces = []
        vxlan_id = 1000
        for interface in self.interfaces:
            priv_ip_lib.create_interface(interface, self.namespace, 'dummy')
            vxlan_interface = interface + '_' + str(vxlan_id)
            vxlan_interfaces.append(vxlan_interface)
            priv_ip_lib.create_interface(
                vxlan_interface, self.namespace, 'vxlan',
                physical_interface=interface, vxlan_id=vxlan_id,
                vxlan_group='239.1.1.1')
            vxlan_id += 1

        devices = priv_ip_lib.get_link_devices(self.namespace)
        self.assertGreater(len(devices), 0)
        device_name_index = {}
        for device in devices:
            name = linux_utils.get_attr(device, 'IFLA_IFNAME')
            device_name_index[name] = device['index']

        for device in devices:
            name = linux_utils.get_attr(device, 'IFLA_IFNAME')
            if name in self.interfaces_to_exclude:
                continue
            self.assertIn(name, self.interfaces + vxlan_interfaces)
            ifla_linkinfo = linux_utils.get_attr(device, 'IFLA_LINKINFO')
            if name in vxlan_interfaces:
                self.assertEqual(
                    linux_utils.get_attr(ifla_linkinfo, 'IFLA_INFO_KIND'),
                    'vxlan')
                ifla_infodata = linux_utils.get_attr(ifla_linkinfo,
                                                     'IFLA_INFO_DATA')
                vxlan_id = int(name.split('_')[-1])
                self.assertEqual(
                    linux_utils.get_attr(ifla_infodata, 'IFLA_VXLAN_ID'),
                    vxlan_id)
                self.assertEqual(
                    linux_utils.get_attr(ifla_infodata, 'IFLA_VXLAN_GROUP'),
                    '239.1.1.1')
                vxlan_link_name = self.interfaces[vxlan_interfaces.index(name)]
                vxlan_link_index = device_name_index[vxlan_link_name]
                self.assertEqual(
                    vxlan_link_index,
                    linux_utils.get_attr(ifla_infodata, 'IFLA_VXLAN_LINK'))
            interfaces_tested.append(name)
        self.assertEqual(sorted(interfaces_tested),
                         sorted(self.interfaces + vxlan_interfaces))

    def _retrieve_interface(self, interface_name, namespace):
        for device in priv_ip_lib.get_link_devices(namespace):
            if interface_name == linux_utils.get_attr(device, 'IFLA_IFNAME'):
                return device
        self.fail('Interface "%s" not found' % interface_name)

    def test_get_link_devices_veth_different_namespaces(self):
        namespace2 = 'ns_test-' + uuidutils.generate_uuid()
        priv_ip_lib.create_netns(namespace2)
        self.addCleanup(self._remove_ns, namespace2)
        # Create a random number of dummy interfaces in namespace2, in order
        # to increase the 'veth1_2' interface index in its namespace.
        for idx in range(5, random.randint(15, 20)):
            priv_ip_lib.create_interface('int_%s' % idx, namespace2, 'dummy')

        ip_wrapper = ip_lib.IPWrapper(self.namespace)
        ip_wrapper.add_veth('veth1_1', 'veth1_2', namespace2)

        veth1_1 = self._retrieve_interface('veth1_1', self.namespace)
        veth1_2 = self._retrieve_interface('veth1_2', namespace2)

        ifla_linkinfo = linux_utils.get_attr(veth1_1, 'IFLA_LINKINFO')
        self.assertEqual(linux_utils.get_attr(ifla_linkinfo, 'IFLA_INFO_KIND'),
                         'veth')
        # NOTE(ralonsoh): since kernel_version=4.15.0-60-generic, iproute2
        # provides the veth pair index, even if the pair interface is in other
        # namespace. In previous versions, the parameter 'IFLA_LINK' was not
        # present. We need to handle both cases.
        self.assertIn(linux_utils.get_attr(veth1_1, 'IFLA_LINK'),
                      [None, veth1_2['index']])

    def test_get_link_devices_veth_same_namespaces(self):
        ip_wrapper = ip_lib.IPWrapper(self.namespace)
        ip_wrapper.add_veth('veth1_1', 'veth1_2')

        veth1_1 = self._retrieve_interface('veth1_1', self.namespace)
        veth1_2 = self._retrieve_interface('veth1_2', self.namespace)

        veth1_1_link = linux_utils.get_attr(veth1_1, 'IFLA_LINK')
        veth1_2_link = linux_utils.get_attr(veth1_2, 'IFLA_LINK')
        self.assertEqual(veth1_1['index'], veth1_2_link)
        self.assertEqual(veth1_2['index'], veth1_1_link)

    def test_get_link_devices_using_index(self):
        for interface in self.interfaces:
            priv_ip_lib.create_interface(interface, self.namespace, 'dummy')
        for expected_device in priv_ip_lib.get_link_devices(self.namespace):
            device = priv_ip_lib.get_link_devices(
                self.namespace, index=expected_device['index'])
            self.assertEqual(1, len(device))
            self.assertEqual(linux_utils.get_attr(expected_device,
                                                  'IFLA_IFNAME'),
                             linux_utils.get_attr(device[0], 'IFLA_IFNAME'))

        self.assertRaises(netlink_exc.NetlinkError,
                          priv_ip_lib.get_link_devices, self.namespace,
                          index=10000)


class BaseIpRuleTestCase(functional_base.BaseSudoTestCase):

    def setUp(self):
        super().setUp()
        self.namespace = 'ns_test-' + uuidutils.generate_uuid()
        self.ns = priv_ip_lib.create_netns(self.namespace)
        self.addCleanup(self._remove_ns)

    def _remove_ns(self):
        priv_ip_lib.remove_netns(self.namespace)

    def _check_rules(self, rules, parameters, values, exception_string=None,
                     raise_exception=True):
        for rule in rules:
            if all(rule.get(parameter) == value
                   for parameter, value in zip(parameters, values)):
                return True
        else:
            if raise_exception:
                self.fail('Rule with %s was expected' % exception_string)
            return False


class ListIpRulesTestCase(BaseIpRuleTestCase):

    RULE_TABLES = {'default': 253, 'main': 254, 'local': 255}

    def test_list_default_rules_ipv4(self):
        rules_ipv4 = priv_ip_lib.list_ip_rules(self.namespace, 4)
        self.assertEqual(3, len(rules_ipv4))
        rule_tables = list(self.RULE_TABLES.values())
        for rule in rules_ipv4:
            rule_tables.remove(rule['table'])
        self.assertEqual(0, len(rule_tables))

    def test_list_rules_ipv4(self):
        ip_lib.add_ip_rule(self.namespace, '192.168.0.1/24', table=10)
        rules_ipv4 = priv_ip_lib.list_ip_rules(self.namespace, 4)
        for rule in rules_ipv4:
            if rule['table'] == 10:
                self.assertEqual('192.168.0.1', rule['attrs']['FRA_SRC'])
                self.assertEqual(24, rule['src_len'])
                break
        else:
            self.fail('Rule added (192.168.0.1/24, table 10) not found')

    def test_list_default_rules_ipv6(self):
        rules_ipv6 = priv_ip_lib.list_ip_rules(self.namespace, 6)
        self.assertEqual(2, len(rules_ipv6))
        rule_tables = [255, 254]
        for rule in rules_ipv6:
            rule_tables.remove(rule['table'])
        self.assertEqual(0, len(rule_tables))

    def test_list_rules_ipv6(self):
        ip_lib.add_ip_rule(self.namespace, '2001:db8::1/64', table=20)
        rules_ipv6 = priv_ip_lib.list_ip_rules(self.namespace, 6)
        for rule in rules_ipv6:
            if rule['table'] == 20:
                self.assertEqual('2001:db8::1', rule['attrs']['FRA_SRC'])
                self.assertEqual(64, rule['src_len'])
                break
        else:
            self.fail('Rule added (2001:db8::1/64, table 20) not found')


class AddIpRulesTestCase(BaseIpRuleTestCase):

    def test_add_rule_ip(self):
        ip_addresses = ['192.168.200.250', '2001::250']
        for ip_address in ip_addresses:
            ip_version = common_utils.get_ip_version(ip_address)
            ip_lenght = common_utils.get_network_length(ip_version)
            ip_family = common_utils.get_socket_address_family(ip_version)
            priv_ip_lib.add_ip_rule(self.namespace, src=ip_address,
                                    src_len=ip_lenght, family=ip_family,
                                    table=ip_lib.IP_RULE_TABLES['default'])
            rules = ip_lib.list_ip_rules(self.namespace, ip_version)
            self._check_rules(rules, ['from'], [ip_address],
                              '"from" IP address %s' % ip_address)

            priv_ip_lib.delete_ip_rule(self.namespace, family=ip_family,
                                       src=ip_address, src_len=ip_lenght,
                                       table=ip_lib.IP_RULE_TABLES['default'])
            rules = ip_lib.list_ip_rules(self.namespace, ip_version)
            self.assertFalse(
                self._check_rules(rules, ['from'], [ip_address],
                                  raise_exception=False))

    def test_add_rule_iif(self):
        iif = 'iif_device_1'
        priv_ip_lib.create_interface(iif, self.namespace, 'dummy')
        priv_ip_lib.add_ip_rule(self.namespace, iifname=iif)
        rules = ip_lib.list_ip_rules(self.namespace, 4)
        self._check_rules(rules, ['iif'], [iif], 'iif name %s' % iif)

        priv_ip_lib.delete_ip_rule(self.namespace, iifname=iif)
        rules = ip_lib.list_ip_rules(self.namespace, 4)
        self.assertFalse(
            self._check_rules(rules, ['iif'], [iif], raise_exception=False))

    def test_add_rule_table(self):
        table = 212
        ip_addresses = ['192.168.200.251', '2001::251']
        for ip_address in ip_addresses:
            ip_version = common_utils.get_ip_version(ip_address)
            ip_lenght = common_utils.get_network_length(ip_version)
            ip_family = common_utils.get_socket_address_family(ip_version)
            priv_ip_lib.add_ip_rule(self.namespace, table=table,
                                    src=ip_address, src_len=ip_lenght,
                                    family=ip_family)
            rules = ip_lib.list_ip_rules(self.namespace, ip_version)
            self._check_rules(
                rules, ['table', 'from'], [str(table), ip_address],
                'table %s and "from" IP address %s' % (table, ip_address))

            priv_ip_lib.delete_ip_rule(self.namespace, table=table,
                                       src=ip_address, src_len=ip_lenght,
                                       family=ip_family)
            rules = ip_lib.list_ip_rules(self.namespace, ip_version)
            self.assertFalse(
                self._check_rules(rules, ['table', 'from'],
                                  [str(table), ip_address],
                                  raise_exception=False))

    def test_add_rule_priority(self):
        priority = 12345
        ip_addresses = ['192.168.200.252', '2001::252']
        for ip_address in ip_addresses:
            ip_version = common_utils.get_ip_version(ip_address)
            ip_lenght = common_utils.get_network_length(ip_version)
            ip_family = common_utils.get_socket_address_family(ip_version)
            priv_ip_lib.add_ip_rule(self.namespace, priority=priority,
                                    src=ip_address, src_len=ip_lenght,
                                    family=ip_family,
                                    table=ip_lib.IP_RULE_TABLES['default'])
            rules = ip_lib.list_ip_rules(self.namespace, ip_version)
            self._check_rules(
                rules, ['priority', 'from'], [str(priority), ip_address],
                'priority %s and "from" IP address %s' %
                (priority, ip_address))

            priv_ip_lib.delete_ip_rule(self.namespace, priority=priority,
                                       src=ip_address, src_len=ip_lenght,
                                       family=ip_family,
                                       table=ip_lib.IP_RULE_TABLES['default'])
            rules = ip_lib.list_ip_rules(self.namespace, ip_version)
            self.assertFalse(
                self._check_rules(rules, ['priority', 'from'],
                                  [str(priority), ip_address],
                                  raise_exception=False))

    def test_add_rule_priority_table_iif(self):
        table = 213
        priority = 12346
        iif = 'iif_device_2'
        priv_ip_lib.create_interface(iif, self.namespace, 'dummy')
        priv_ip_lib.add_ip_rule(self.namespace, priority=priority, iifname=iif,
                                table=table)

        rules = ip_lib.list_ip_rules(self.namespace, 4)
        self._check_rules(
            rules, ['priority', 'iif', 'table'],
            [str(priority), iif, str(table)],
            'priority %s, table %s and iif name %s' % (priority, table, iif))

        priv_ip_lib.delete_ip_rule(self.namespace, priority=priority,
                                   iifname=iif, table=table)
        rules = ip_lib.list_ip_rules(self.namespace, 4)
        self.assertFalse(
            self._check_rules(rules, ['priority', 'iif', 'table'],
                              [str(priority), iif, str(table)],
                              raise_exception=False))

    @testtools.skip('https://github.com/svinota/pyroute2/issues/566')
    def test_add_rule_exists(self):
        iif = 'iif_device_1'
        priv_ip_lib.create_interface(iif, self.namespace, 'dummy')
        priv_ip_lib.add_ip_rule(self.namespace, iifname=iif,
                                table=ip_lib.IP_RULE_TABLES['default'])
        rules = ip_lib.list_ip_rules(self.namespace, 4)
        self._check_rules(rules, ['iif'], [iif], 'iif name %s' % iif)
        self.assertEqual(4, len(rules))

        # pyroute2.netlink.exceptions.NetlinkError(17, 'File exists')
        # exception is catch.
        priv_ip_lib.add_ip_rule(self.namespace, iifname=iif,
                                table=ip_lib.IP_RULE_TABLES['default'])
        rules = ip_lib.list_ip_rules(self.namespace, 4)
        self._check_rules(rules, ['iif'], [iif], 'iif name %s' % iif)
        self.assertEqual(4, len(rules))


class DeleteIpRulesTestCase(BaseIpRuleTestCase):

    def test_delete_rule_no_entry(self):
        iif = 'iif_device'
        priv_ip_lib.create_interface(iif, self.namespace, 'dummy')

        try:
            # This should not raise for a non-existent entry
            priv_ip_lib.delete_ip_rule(self.namespace, iifname=iif)
        except Exception:
            self.fail('Delete IP rule threw unexpected exception')

        rules = ip_lib.list_ip_rules(self.namespace, 4)
        # There are always 3 rules by default
        self.assertEqual(3, len(rules))


class GetIpAddressesTestCase(functional_base.BaseSudoTestCase):

    def _remove_ns(self, namespace):
        priv_ip_lib.remove_netns(namespace)

    def test_get_ip_addresses(self):
        namespace = 'ns_test-' + uuidutils.generate_uuid()
        priv_ip_lib.create_netns(namespace)
        self.addCleanup(self._remove_ns, namespace)
        interfaces = {
            '20': {'cidr': '192.168.10.20/24', 'scope': 'link',
                   'add_broadcast': True},
            '30': {'cidr': '2001::1/64', 'scope': 'global',
                   'add_broadcast': False}}

        for int_name, int_parameters in interfaces.items():
            priv_ip_lib.create_interface(int_name, namespace, 'dummy',
                                         index=int(int_name))
            ip_lib.add_ip_address(
                int_parameters['cidr'], int_name, namespace,
                int_parameters['scope'], int_parameters['add_broadcast'])

        ip_addresses = priv_ip_lib.get_ip_addresses(namespace)
        for ip_address in ip_addresses:
            int_name = str(ip_address['index'])
            ip = linux_utils.get_attr(ip_address, 'IFA_ADDRESS')
            mask = ip_address['prefixlen']
            cidr = common_utils.ip_to_cidr(ip, mask)
            self.assertEqual(interfaces[int_name]['cidr'], cidr)
            self.assertEqual(interfaces[int_name]['scope'],
                             ip_lib.IP_ADDRESS_SCOPE[ip_address['scope']])


class RouteTestCase(functional_base.BaseSudoTestCase):

    def setUp(self):
        super(RouteTestCase, self).setUp()
        self.namespace = self.useFixture(net_helpers.NamespaceFixture()).name
        self.device_name = 'test_device'
        ip_lib.IPWrapper(self.namespace).add_dummy(self.device_name)
        self.device = ip_lib.IPDevice(self.device_name, self.namespace)
        self.device.link.set_up()

    def _check_gateway_or_multipath(self, route, gateway):
        if gateway is None or isinstance(gateway, str):
            self.assertEqual(gateway,
                            linux_utils.get_attr(route, 'RTA_GATEWAY'))
            return

        rta_multipath = linux_utils.get_attr(route, 'RTA_MULTIPATH')
        self.assertEqual(len(gateway), len(rta_multipath))
        for nexthop in gateway:
            to_check = {'hops': 0}
            if nexthop.get('device'):
                to_check['oif'] = priv_ip_lib.privileged_get_link_id(
                    nexthop['device'], self.namespace)
            if nexthop.get('weight'):
                to_check['hops'] = nexthop['weight'] - 1
            if nexthop.get('via'):
                to_check['gateway'] = nexthop['via']

            for mp in rta_multipath:
                mp['gateway'] = linux_utils.get_attr(mp, 'RTA_GATEWAY')
                for key in to_check:
                    if to_check[key] != mp[key]:
                        break
                else:
                    break
            else:
                self.fail('Route nexthops %s do not match with defined ones '
                          '%s' % (rta_multipath, gateway))

    def _check_routes(self, cidrs, table=None, gateway=None, metric=None,
                      scope=None, proto='static'):
        table = table or iproute_linux.DEFAULT_TABLE
        if not scope:
            scope = 'universe' if gateway else 'link'
        scope = priv_ip_lib.get_scope_name(scope)
        for cidr in cidrs:
            ip_version = common_utils.get_ip_version(cidr)
            if ip_version == n_cons.IP_VERSION_6 and not metric:
                metric = ipdb_routes.IP6_RT_PRIO_USER
            if ip_version == n_cons.IP_VERSION_6:
                scope = 0
            routes = priv_ip_lib.list_ip_routes(self.namespace, ip_version)
            for route in routes:
                ip = linux_utils.get_attr(route, 'RTA_DST')
                mask = route['dst_len']
                if not (ip == str(netaddr.IPNetwork(cidr).ip) and
                        mask == netaddr.IPNetwork(cidr).cidr.prefixlen):
                    continue
                self.assertEqual(table, route['table'])
                self.assertEqual(
                    priv_ip_lib._IP_VERSION_FAMILY_MAP[ip_version],
                    route['family'])
                self._check_gateway_or_multipath(route, gateway)
                self.assertEqual(metric,
                                 linux_utils.get_attr(route, 'RTA_PRIORITY'))
                self.assertEqual(scope, route['scope'])
                self.assertEqual(rtnl.rt_proto[proto], route['proto'])
                break
            else:
                self.fail('CIDR %s not found in the list of routes' % cidr)

    def _check_gateway(self, gateway, table=None, metric=None):
        table = table or iproute_linux.DEFAULT_TABLE
        ip_version = common_utils.get_ip_version(gateway)
        if ip_version == n_cons.IP_VERSION_6 and not metric:
            metric = ipdb_routes.IP6_RT_PRIO_USER
        scope = 0
        routes = priv_ip_lib.list_ip_routes(self.namespace, ip_version)
        for route in routes:
            if not (linux_utils.get_attr(route, 'RTA_GATEWAY') == gateway):
                continue
            self.assertEqual(table, route['table'])
            self.assertEqual(
                priv_ip_lib._IP_VERSION_FAMILY_MAP[ip_version],
                route['family'])
            self.assertEqual(gateway,
                             linux_utils.get_attr(route, 'RTA_GATEWAY'))
            self.assertEqual(scope, route['scope'])
            self.assertEqual(0, route['dst_len'])
            self.assertEqual(metric,
                             linux_utils.get_attr(route, 'RTA_PRIORITY'))
            break
        else:
            self.fail('Default gateway %s not found in the list of routes'
                      % gateway)

    def _add_route_device_and_check(self, table=None, metric=None,
                                    scope='link', proto='static'):
        cidrs = ['192.168.0.0/24', '172.90.0.0/16', '10.0.0.0/8',
                 '2001:db8::/64']
        for cidr in cidrs:
            ip_version = common_utils.get_ip_version(cidr)
            priv_ip_lib.add_ip_route(self.namespace, cidr, ip_version,
                                     device=self.device_name, table=table,
                                     metric=metric, scope=scope, proto=proto)

        self._check_routes(cidrs, table=table, metric=metric, scope=scope,
                           proto=proto)

    def test_add_route_device(self):
        self._add_route_device_and_check(table=None)

    def test_add_route_device_table(self):
        self._add_route_device_and_check(table=100)

    def test_add_route_device_metric(self):
        self._add_route_device_and_check(metric=50)

    def test_add_route_device_table_metric(self):
        self._add_route_device_and_check(table=200, metric=30)

    def test_add_route_device_scope_global(self):
        self._add_route_device_and_check(scope='global')

    def test_add_route_device_scope_site(self):
        self._add_route_device_and_check(scope='site')

    def test_add_route_device_scope_host(self):
        self._add_route_device_and_check(scope='host')

    def test_add_route_device_proto_static(self):
        self._add_route_device_and_check(proto='static')  # default

    def test_add_route_device_proto_redirect(self):
        self._add_route_device_and_check(proto='redirect')

    def test_add_route_device_proto_kernel(self):
        self._add_route_device_and_check(proto='kernel')

    def test_add_route_device_proto_boot(self):
        self._add_route_device_and_check(proto='boot')

    def test_add_route_via_ipv4(self):
        cidrs = ['192.168.0.0/24', '172.90.0.0/16', '10.0.0.0/8']
        int_cidr = '192.168.20.1/24'
        int_ip_address = str(netaddr.IPNetwork(int_cidr).ip)
        self.device.addr.add(int_cidr)
        for cidr in cidrs:
            ip_version = common_utils.get_ip_version(cidr)
            priv_ip_lib.add_ip_route(self.namespace, cidr, ip_version,
                                     via=int_ip_address)
        self._check_routes(cidrs, gateway=int_ip_address)

    def test_add_route_via_ipv6(self):
        cidrs = ['2001:db8::/64', 'faaa::/96']
        int_cidr = 'fd00::1/64'
        via_ip = 'fd00::2'
        self.device.addr.add(int_cidr)
        for cidr in cidrs:
            ip_version = common_utils.get_ip_version(cidr)
            priv_ip_lib.add_ip_route(self.namespace, cidr, ip_version,
                                     via=via_ip)
        self._check_routes(cidrs, gateway=via_ip)

    def test_add_default(self):
        ip_addresses = ['192.168.0.1/24', '172.90.0.1/16', '10.0.0.1/8',
                        '2001:db8::1/64', 'faaa::1/96']
        for ip_address in ip_addresses:
            ip_version = common_utils.get_ip_version(ip_address)
            if ip_version == n_cons.IP_VERSION_4:
                _ip = str(netaddr.IPNetwork(ip_address).ip)
            else:
                _ip = str(netaddr.IPNetwork(ip_address).ip + 1)
            self.device.addr.add(ip_address)
            priv_ip_lib.add_ip_route(self.namespace, None, ip_version,
                                     device=self.device_name, via=_ip)
            self._check_gateway(_ip)

    def test_add_multipath_route(self):
        self.device_name2 = 'test_device2'
        ip_lib.IPWrapper(self.namespace).add_dummy(self.device_name2)
        self.device2 = ip_lib.IPDevice(self.device_name2, self.namespace)
        self.device2.link.set_up()
        self.device.addr.add('10.1.0.1/24')
        self.device2.addr.add('10.2.0.1/24')
        multipath = [
            {'device': self.device_name, 'via': '10.1.0.100', 'weight': 10},
            {'device': self.device_name2, 'via': '10.2.0.100', 'weight': 20},
            {'via': '10.2.0.101', 'weight': 30},
            {'via': '10.2.0.102'}]
        priv_ip_lib.add_ip_route(self.namespace, '192.168.0.0/24',
                                 n_cons.IP_VERSION_4, via=multipath)
        self._check_routes(['192.168.0.0/24'], gateway=multipath)

    def test_delete_route_no_entry(self):
        cidr = '192.168.0.0/24'
        self.device.addr.add('10.1.0.1/24')
        try:
            # This should not raise for a non-existent entry
            priv_ip_lib.delete_ip_route(self.namespace, cidr,
                                        n_cons.IP_VERSION_4,
                                        device=self.device_name)
        except Exception:
            self.fail('Delete IP route threw unexpected exception')

        routes = ip_lib.list_ip_routes(self.namespace, n_cons.IP_VERSION_4)
        # There will be a single interface route since we added an IP
        self.assertEqual(1, len(routes))


class GetLinkAttributesTestCase(functional_base.BaseSudoTestCase):

    def setUp(self):
        super(GetLinkAttributesTestCase, self).setUp()
        self.namespace = self.useFixture(net_helpers.NamespaceFixture()).name
        self.device_name = 'test_device'
        ip_lib.IPWrapper(self.namespace).add_dummy(self.device_name)
        self.device = ip_lib.IPDevice(self.device_name, self.namespace)
        self.pyroute_dev = priv_ip_lib.get_link_devices(
            self.namespace, ifname=self.device_name)
        self.assertEqual(1, len(self.pyroute_dev))
        self.pyroute_dev = self.pyroute_dev[0]

    def test_get_link_attribute_kind(self):
        ifla_linkinfo = linux_utils.get_attr(self.pyroute_dev, 'IFLA_LINKINFO')
        ifla_link_kind = linux_utils.get_attr(ifla_linkinfo, 'IFLA_INFO_KIND')
        self.assertEqual('dummy', ifla_link_kind)
        self.assertEqual(ifla_link_kind, self.device.link.link_kind)

    def test_get_link_attributes(self):
        expected_attr = ['mtu', 'qlen', 'state', 'qdisc', 'brd', 'link/ether',
                         'alias', 'allmulticast', 'link_kind']
        attr = self.device.link.attributes
        self.assertSetEqual(set(expected_attr), set(attr.keys()))


class ListNamespacePids(functional_base.BaseSudoTestCase):

    def setUp(self):
        super(ListNamespacePids, self).setUp()
        self.namespace = self.useFixture(net_helpers.NamespaceFixture()).name
        self.timeout = 3

    @staticmethod
    def _run_sleep(namespace, timeout):
        ip_wrapper = ip_lib.IPWrapper(namespace=namespace)
        # NOTE(ralonsoh): this is a "long" (more than one second) lived
        # process. It should not be executed in a privsep context to avoid
        # a possible privsep thread starvation.
        ip_wrapper.netns.execute(['sleep', timeout], check_exit_code=False)

    def _check_pids(self, num_pids, namespace=None):
        namespace = self.namespace if not namespace else namespace
        self.pids = priv_ip_lib.list_ns_pids(namespace)
        return len(self.pids) == num_pids

    def test_list_namespace_pids(self):
        thread = threading.Thread(target=self._run_sleep,
                                  args=(self.namespace, self.timeout))
        thread.start()
        try:
            check_pids = functools.partial(self._check_pids, 1)
            common_utils.wait_until_true(check_pids, timeout=self.timeout)
        except common_utils.WaitTimeout:
            self.fail('Process no found in namespace %s' % self.namespace)
        thread.join(timeout=self.timeout)

    def test_list_namespace_pids_nothing_running_inside(self):
        self.assertTrue(self._check_pids(0))

    def test_list_namespace_not_created(self):
        self.assertTrue(self._check_pids(0, namespace='othernamespace'))


class GetLinkIdTestCase(functional_base.BaseSudoTestCase):

    def _remove_device(self, device_name):
        priv_ip_lib.delete_interface(device_name, None)

    def test_get_link_id_device_found(self):
        device_name = 'dev_' + uuidutils.generate_uuid()[:11]
        ip_lib.IPWrapper().add_dummy(device_name)
        ip_lib.IPDevice(device_name)
        self.addCleanup(self._remove_device, device_name)
        self.assertGreater(priv_ip_lib.get_link_id(device_name, None), 0)

    def test_get_link_id_device_not_found_raise_exception(self):
        device_name = 'strange_device'
        self.assertRaises(priv_ip_lib.NetworkInterfaceNotFound,
                          priv_ip_lib.get_link_id, device_name, None)

    def test_get_link_id_device_not_found_do_not_raise_exception(self):
        device_name = 'strange_device'
        with mock.patch.object(priv_ip_lib, 'LOG') as mock_log:
            priv_ip_lib.get_link_id(device_name, None, raise_exception=False)
        mock_log.debug.assert_called_once_with(
            'Interface %(dev)s not found in namespace %(namespace)s',
            {'dev': device_name, 'namespace': None})



# Copyright (c) 2015 OpenStack Foundation.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from neutron_lib.agent import topics
from neutron_lib import constants
from oslo_config import cfg
from oslo_service import service

from neutron.common import config as common_config
from neutron.conf.agent import common as config
from neutron.conf.agent.l3 import config as l3_config
from neutron.conf.agent.l3 import ha as ha_conf
from neutron.conf.agent.metadata import config as meta_conf
from neutron.conf.plugins.ml2.drivers import ovs_conf
from neutron import service as neutron_service


def register_opts(conf):
    common_config.register_common_config_options()
    l3_config.register_l3_agent_config_opts(l3_config.OPTS, conf)
    ha_conf.register_l3_agent_ha_opts(conf)
    meta_conf.register_meta_conf_opts(meta_conf.SHARED_OPTS, conf)
    meta_conf.register_meta_conf_opts(meta_conf.METADATA_RATE_LIMITING_OPTS,
                                      cfg=conf,
                                      group=meta_conf.RATE_LIMITING_GROUP)
    config.register_interface_driver_opts_helper(conf)
    config.register_agent_state_opts_helper(conf)
    config.register_interface_opts(conf)
    config.register_external_process_opts(conf)
    config.register_pddriver_opts(conf)
    config.register_ra_opts(conf)
    config.register_availability_zone_opts_helper(conf)
    ovs_conf.register_ovs_opts(conf)


def main(manager='neutron.agent.l3.agent.L3NATAgentWithStateReport'):
    register_opts(cfg.CONF)
    common_config.init(sys.argv[1:])
    config.setup_logging()
    config.setup_privsep()
    server = neutron_service.Service.create(
        binary=constants.AGENT_PROCESS_L3,
        topic=topics.L3_AGENT,
        report_interval=cfg.CONF.AGENT.report_interval,
        manager=manager)
    service.launch(cfg.CONF, server, restart_method='mutate').wait()



# Copyright 2017 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import grp
import os
import pwd

from neutron.agent.linux import external_process
from neutron.agent.linux import ip_lib
from neutron_lib import exceptions
from oslo_config import cfg
from oslo_log import log as logging

from neutron._i18n import _
from neutron.common import metadata as comm_meta

LOG = logging.getLogger(__name__)

METADATA_SERVICE_NAME = 'metadata-proxy'
HAPROXY_SERVICE = 'haproxy'

PROXY_CONFIG_DIR = "ovn-metadata-proxy"

_HEADER_CONFIG_TEMPLATE = """
    http-request add-header X-OVN-%(res_type)s-ID %(res_id)s
"""

_UNLIMITED_CONFIG_TEMPLATE = """
listen listener
    bind %(host)s:%(port)s
    %(bind_v6_line)s
    server metadata %(unix_socket_path)s
"""


class HaproxyConfigurator(object):
    def __init__(self, network_id, router_id, unix_socket_path, host,
                 port, user, group, state_path, pid_file,
                 rate_limiting_config, host_v6=None,
                 bind_interface=None):
        self.network_id = network_id
        self.router_id = router_id
        if network_id is None and router_id is None:
            raise exceptions.NetworkIdOrRouterIdRequiredError()

        self.host = host
        self.host_v6 = host_v6
        self.bind_interface = bind_interface
        self.port = port
        self.user = user
        self.group = group
        self.state_path = state_path
        self.unix_socket_path = unix_socket_path
        self.pidfile = pid_file
        self.rate_limiting_config = rate_limiting_config
        self.log_level = (
            'debug' if logging.is_debug_enabled(cfg.CONF) else 'info')
        # log-tag will cause entries to have the string pre-pended, so use
        # the uuid haproxy will be started with.  Additionally, if it
        # starts with "haproxy" then things will get logged to
        # /var/log/haproxy.log on Debian distros, instead of to syslog.
        uuid = network_id or router_id
        self.log_tag = 'haproxy-{}-{}'.format(METADATA_SERVICE_NAME, uuid)

    def create_config_file(self):
        """Create the config file for haproxy."""
        # Need to convert uid/gid into username/group
        try:
            username = pwd.getpwuid(int(self.user)).pw_name
        except (ValueError, KeyError):
            try:
                username = pwd.getpwnam(self.user).pw_name
            except KeyError:
                raise comm_meta.InvalidUserOrGroupException(
                    _("Invalid user/uid: '%s'") % self.user)

        try:
            groupname = grp.getgrgid(int(self.group)).gr_name
        except (ValueError, KeyError):
            try:
                groupname = grp.getgrnam(self.group).gr_name
            except KeyError:
                raise comm_meta.InvalidUserOrGroupException(
                    _("Invalid group/gid: '%s'") % self.group)

        cfg_info = {
            'host': self.host,
            'port': self.port,
            'unix_socket_path': self.unix_socket_path,
            'user': username,
            'group': groupname,
            'pidfile': self.pidfile,
            'log_level': self.log_level,
            'log_tag': self.log_tag,
            'bind_v6_line': '',
        }
        if self.host_v6 and self.bind_interface:
            cfg_info['bind_v6_line'] = (
                'bind %s:%s interface %s' % (
                    self.host_v6, self.port, self.bind_interface)
            )
        if self.network_id:
            cfg_info['res_type'] = 'Network'
            cfg_info['res_id'] = self.network_id
        else:
            cfg_info['res_type'] = 'Router'
            cfg_info['res_id'] = self.router_id

        haproxy_cfg = comm_meta.get_haproxy_config(cfg_info,
                                                   self.rate_limiting_config,
                                                   _HEADER_CONFIG_TEMPLATE,
                                                   _UNLIMITED_CONFIG_TEMPLATE)

        LOG.debug("haproxy_cfg = %s", haproxy_cfg)
        cfg_dir = self.get_config_path(self.state_path)
        # uuid has to be included somewhere in the command line so that it can
        # be tracked by process_monitor.
        self.cfg_path = os.path.join(cfg_dir, "%s.conf" % cfg_info['res_id'])
        if not os.path.exists(cfg_dir):
            os.makedirs(cfg_dir)
        with open(self.cfg_path, "w") as cfg_file:
            cfg_file.write(haproxy_cfg)

    @staticmethod
    def get_config_path(state_path):
        return os.path.join(state_path or cfg.CONF.state_path,
                            PROXY_CONFIG_DIR)

    @staticmethod
    def cleanup_config_file(uuid, state_path):
        """Delete config file created when metadata proxy was spawned."""
        # Delete config file if it exists
        cfg_path = os.path.join(
            HaproxyConfigurator.get_config_path(state_path),
            "%s.conf" % uuid)
        try:
            os.unlink(cfg_path)
        except OSError as ex:
            # It can happen that this function is called but metadata proxy
            # was never spawned so its config file won't exist
            if ex.errno != errno.ENOENT:
                raise


class MetadataDriver(object):

    monitors = {}

    @classmethod
    def _get_metadata_proxy_user_group(cls, conf):
        user = conf.metadata_proxy_user or str(os.geteuid())
        group = conf.metadata_proxy_group or str(os.getegid())

        return user, group

    @classmethod
    def _get_metadata_proxy_callback(cls, bind_address, port, conf,
                                     network_id=None, router_id=None,
                                     bind_address_v6=None,
                                     bind_interface=None):
        def callback(pid_file):
            metadata_proxy_socket = conf.metadata_proxy_socket
            user, group = (
                cls._get_metadata_proxy_user_group(conf))
            haproxy = HaproxyConfigurator(network_id,
                                          router_id,
                                          metadata_proxy_socket,
                                          bind_address,
                                          port,
                                          user,
                                          group,
                                          conf.state_path,
                                          pid_file,
                                          conf.metadata_rate_limiting,
                                          bind_address_v6,
                                          bind_interface)
            haproxy.create_config_file()
            proxy_cmd = [HAPROXY_SERVICE,
                         '-f', haproxy.cfg_path]
            return proxy_cmd

        return callback

    @classmethod
    def spawn_monitored_metadata_proxy(cls, monitor, ns_name, port, conf,
                                       bind_address="0.0.0.0", network_id=None,
                                       router_id=None, bind_address_v6=None,
                                       bind_interface=None):
        uuid = network_id or router_id
        callback = cls._get_metadata_proxy_callback(
            bind_address, port, conf, network_id=network_id,
            router_id=router_id, bind_address_v6=bind_address_v6,
            bind_interface=bind_interface)
        pm = cls._get_metadata_proxy_process_manager(uuid, conf,
                                                     ns_name=ns_name,
                                                     callback=callback)
        if bind_interface is not None and bind_address_v6 is not None:
            # HAProxy cannot bind() until IPv6 Duplicate Address Detection
            # completes. We must wait until the address leaves its 'tentative'
            # state.
            ip_lib.IpAddrCommand(
                parent=ip_lib.IPDevice(name=bind_interface, namespace=ns_name)
            ).wait_until_address_ready(address=bind_address_v6)
        try:
            pm.enable()
        except exceptions.ProcessExecutionError as exec_err:
            LOG.error("Encountered process execution error %(err)s while "
                      "starting process in namespace %(ns)s",
                      {"err": exec_err, "ns": ns_name})
            return
        monitor.register(uuid, METADATA_SERVICE_NAME, pm)
        cls.monitors[router_id] = pm

    @classmethod
    def destroy_monitored_metadata_proxy(cls, monitor, uuid, conf, ns_name):
        monitor.unregister(uuid, METADATA_SERVICE_NAME)
        pm = cls._get_metadata_proxy_process_manager(uuid, conf,
                                                     ns_name=ns_name)
        pm.disable()

        # Delete metadata proxy config file
        HaproxyConfigurator.cleanup_config_file(uuid, cfg.CONF.state_path)

        cls.monitors.pop(uuid, None)

    @classmethod
    def _get_metadata_proxy_process_manager(cls, router_id, conf, ns_name=None,
                                            callback=None):
        return external_process.ProcessManager(
            conf=conf,
            uuid=router_id,
            namespace=ns_name,
            service=HAPROXY_SERVICE,
            default_cmd_callback=callback)
